/*
	ProWesS, PROGS Windowing System

	Outline object "OUTL"

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 21 February 1995
	26.10.1995 added two empty direction objects at the left and right inside
			   the outline to allow the client to add extra object in them.
	28.10.1995 make sure outline doesn't crash when SLEEP item asked twice
	06.01.1996 full support for 'help' included
	07.05.1996 added queries for object id's for all loose items (for sbasic if)
	23.04.1997 PW_OBJECT_POINTER_START on Abort in quit confirmation window
*/

#include <mem.h>
#include <win.h>
#include <job.h>
#include <thing.h>

#include "ProWesS.h"
#include "PWTypeDef.h"
#include "PWHandler.h"

/* global statusses */

/* object definition */
typedef struct {
	PWBaseObject;					/* common start for all objects */
	PWObject title, sep, tc;		/* always present */
	PWObject contleft, contright;	/* always present - extra obj containers */
	PWObject button, sleep, info;	/* optional */
	PWObject quit, doit, wake, help;/* optional */
	char jobname[JOB_MAXNAME];		/* job name, default title */
	char helpfile[IO_MAXFULLNAME];
	char helpdir[IO_MAXDIRECTORY];
	char helppos[IO_MAXFULLNAME];
} Object;

extern PWTypeDef init;

#define catch(x) do { if (err=(x)) return err; } while (0)

/* object handling routines */

Error Create(PWObject owner, PWObject *ret, PWEvent event, int *taglist)
{
	event.region->composite=TRUE;
	/* allocate specific object - which should be cleared (automatic) */
	return PWAllocate(event.region->system,sizeof(Object),ret);
	/* no events to react to */
}

Error Remove(PWObject object)
{
	Object *obj=(Object *)object;

	if (obj->button) PWRemove(obj->button);

	/* all the other objects are children and are thus automatically removed */

	return PWRelease((PWObject)obj,(PWObject)obj);
}

Error Query(PWObject object, int what, void *ret)
{
	Error err=ERR_OK;
	Object *obj=(Object *)object;

	switch (what)
	{
		case PW_OUTLINE_BOX_LEFT:
			*((PWObject *)ret)=obj->contleft;
			break;
		case PW_OUTLINE_BOX_RIGHT:
			*((PWObject *)ret)=obj->contright;
			break;
		case PW_OUTLINE_OBJECT_QUIT:
			*((PWObject *)ret)=obj->quit;
			break;
		case PW_OUTLINE_OBJECT_INFO:
			*((PWObject *)ret)=obj->info;
			break;
		case PW_OUTLINE_OBJECT_DO:
			*((PWObject *)ret)=obj->doit;
			break;
		case PW_OUTLINE_OBJECT_WAKE:
			*((PWObject *)ret)=obj->wake;
			break;
		case PW_OUTLINE_OBJECT_HELP:
			*((PWObject *)ret)=obj->help;
			break;
		default :
			err=ERR_IPAR;
	}
	return err;
}

static Error sleep(PWObject obj)
{
	return PWChange(obj, PW_SYSTEM_SLEEP, NULL);
}

static Error breakdown(PWObject obj)
{
	return PWChange(obj, PW_SYSTEM_BREAKDOWN, NULL);
}

static Error quitprg(PWObject obj)
{
	int *result;

	PWQuery(obj, PW_OBJECT_AUXILIARY, &result);
	*result=TRUE;
	return PWChange(obj, PW_SYSTEM_BREAKDOWN, NULL);
}

static Error quit(PWObject obj)
{
	PWObject win,yes,no;
	int result=FALSE;

	if (!obj->auxiliary)
		return PWChange(obj, PW_SYSTEM_BREAKDOWN, NULL);
	else
	{
		/* confirmation request */
		PWCreate(NULL,&win,PW_TYPE_OUTLINE,
					PW_TITLE_TEXT, "Are you sure you want to quit ?",
					NULL);
		PWCreate(win,&yes,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "Quit program",
					PW_OBJECT_AUXILIARY, &result,
					PW_LOOSE_ACTION_HIT, &quitprg,
					PW_LOOSE_KEYPRESS, 'q',
					NULL);
		PWCreate(win,&no,PW_TYPE_LOOSE_ITEM,
					PW_LOOSE_TEXT, "Abort",
					PW_OBJECT_AUXILIARY, &result,
					PW_LOOSE_ACTION_HIT, &breakdown,
					PW_LOOSE_KEYPRESS, 27,
					PW_OBJECT_POINTER_START,
					NULL);
		PWActivate(win);
		PWRemove(win);
		if (result) return PWChange(obj, PW_SYSTEM_BREAKDOWN, NULL);
	}
	return ERR_OK;
}

static Error help(PWObject object)
{
	Object *obj;
	char *strings[7]; char **str=strings;

	PWQuery(object,PW_OBJECT_AUXILIARY,&obj);
	*str++="-help";
	*str++="-file";
	*str++=obj->helpfile;
	if (obj->helpdir[0]) *str++="-dir", *str++=obj->helpdir;
	if (obj->helppos[0]) *str++="-pos", *str++=obj->helppos;
	THINGExec("ProWesS reader",8,MYSELF,FALSE,str-strings,strings,0,NULL,NULL);
	return ERR_OK;
}

Error Change(PWObject object, int *taglist)
{
	Object *obj=(Object *)object;
	int op;
	PWSystem *system=obj->event.region->system;
	Error err=ERR_OK;
	char unrecognised_tag=FALSE;

	if (!(obj->status&OSTATUS_INITIALISED))
	{
		obj->status|=OSTATUS_INITIALISED;
		/* create title, container and sleep objects */
		catch( PWCreate((PWObject)obj,&obj->tc,PW_TYPE_DIRECTION,
						PW_SCALE_FACTOR, 0,
						NULL) );
		JOBInfo(MYSELF,obj->jobname,NULL,NULL);
		catch( PWCreate(obj->tc,&obj->title,PW_TYPE_TITLE_ITEM,
						PW_TITLE_TEXT, obj->jobname,
						NULL) );
		catch( PWCreate((PWObject)obj,&obj->sep,PW_TYPE_SEPARATOR,
						PW_POSITION_BELOW, obj->tc,
						NULL) );
		catch( PWCreate(obj->tc,&obj->contleft,PW_TYPE_DIRECTION,
						PW_POSITION_LEFT_OF, obj->title,
						PW_SCALE_FACTOR, 0,
						NULL) );
		catch( PWCreate(obj->tc,&obj->contright,PW_TYPE_DIRECTION,
						PW_POSITION_RIGHT_OF, obj->title,
						PW_SCALE_FACTOR, 0,
						NULL) );
	}

	while (*taglist)
	{
		switch (*taglist)
		{
		case PW_OUTLINE_SLEEP:
			if (obj->sleep) break;	/* there may already be a sleep item */
			catch( PWCreate((PWObject)obj,&obj->sleep,PW_TYPE_LOOSE_ITEM,
							PW_POSITION_RIGHT_OF, obj->quit?obj->quit:obj->title,
							PW_LOOSE_KEYPRESS, 233,
							PW_LOOSE_TEXT, "Zzz",
							PW_LOOSE_CHANGE_STATUS, FALSE,
							PW_LOOSE_ACTION_HIT, &sleep,
							PW_SCALE_FACTOR, 0,
							NULL) );
			catch( PWCreate((PWObject)obj,&obj->button,PW_TYPE_LOOSE_ITEM,
							PW_LOOSE_TEXT, obj->jobname,
							PW_LOOSE_CHANGE_STATUS, FALSE,
							PW_LOOSE_KEYPRESS, KEYSTROKE_WAKE,
/*							  PW_LOOSE_KEYPRESS, 237,			*/
							PW_LOOSE_ACTION_HIT, &breakdown,
							PW_LOOSE_ACTION_DO, &breakdown,
							PW_SLEEP_OBJECT,
							NULL) );
			break;
		case PW_OUTLINE_SLEEP_TEXT:
			if (!obj->button)
			{
				static const int list[2]={PW_OUTLINE_SLEEP,NUL};
				catch( Change(object,list) );
			}
			catch( PWChange(obj->button,PW_LOOSE_TEXT,*(taglist+1),NULL) );
			break;

		case PW_OUTLINE_QUIT:
			catch( PWCreate((PWObject)obj,&obj->quit,PW_TYPE_LOOSE_ITEM,
							PW_POSITION_RIGHT_OF, obj->title,
							PW_LOOSE_TEXT, "quit",
							PW_LOOSE_CHANGE_STATUS, FALSE,
							PW_LOOSE_ACTION_HIT, &quit,
							PW_OBJECT_AUXILIARY, FALSE,
							PW_SCALE_FACTOR, 0,
							NULL) );
			break;
		case PW_OUTLINE_ACTION_QUIT:
			op=PW_LOOSE_ACTION_HIT;
			goto setquititem;
		case PW_OUTLINE_QUIT_CONFIRM:
			op=PW_OBJECT_AUXILIARY;
			goto setquititem;
		case PW_OUTLINE_QUIT_KEYPRESS:
			op=PW_LOOSE_KEYPRESS;
setquititem:
			if (obj->quit==NULL) return ERR_IPAR;
			PWChange(obj->quit,op,*(taglist+1),NULL);
			break;

		case PW_OUTLINE_INFO_TEXT:
			if (obj->info==NULL)
				catch( PWCreate((PWObject)obj,&obj->info,PW_TYPE_LOOSE_ITEM,
								PW_POSITION_LEFT_OF, obj->title,
								PW_LOOSE_TEXT, *(taglist+1),
								PW_LOOSE_CHANGE_STATUS, FALSE,
								PW_SCALE_FACTOR, 0,
								NULL) );
			else
				catch( PWChange(obj->info,PW_LOOSE_TEXT,*(taglist+1),NULL) );
			break;
		case PW_OUTLINE_ACTION_INFO:
			if (obj->info==NULL)
				catch( PWCreate((PWObject)obj,&obj->info,PW_TYPE_LOOSE_ITEM,
								PW_POSITION_LEFT_OF, obj->title,
								PW_LOOSE_TEXT, "info",
								PW_LOOSE_ACTION_HIT, *(taglist+1),
								PW_LOOSE_CHANGE_STATUS, FALSE,
								PW_SCALE_FACTOR, 0,
								NULL) );
			else
				catch( PWChange(obj->info,PW_LOOSE_ACTION_HIT,*(taglist+1),NULL) );
			break;

		case PW_OUTLINE_ACTION_DO:
			if (obj->doit && *(taglist+1)==NUL)
				catch( PWRemove(obj->doit) );
			else if (obj->doit && *(taglist+1))
				catch( PWChange(obj->doit, PW_LOOSE_ACTION_HIT, *(taglist+1), NULL) );
			else if (obj->doit==NULL && *(taglist+1))
				catch( PWCreate((PWObject)obj,&obj->doit,PW_TYPE_LOOSE_ITEM,
								PW_POSITION_LEFT_OF, (obj->help?obj->help:(obj->info?obj->info:obj->title)),
								PW_LOOSE_TEXT, "do",
								PW_LOOSE_KEYPRESS, KEYSTROKE_DO,
								PW_LOOSE_ACTION_HIT, *(taglist+1),
								PW_LOOSE_CHANGE_STATUS, FALSE,
								PW_SCALE_FACTOR, 0,
								NULL) );
			break;

		case PW_OUTLINE_ACTION_WAKE:
			if (obj->wake && *(taglist+1)==NUL)
				catch( PWRemove(obj->wake) );
			else if (obj->wake && *(taglist+1))
				catch( PWChange(obj->wake, PW_LOOSE_ACTION_HIT, *(taglist+1), NULL) );
			else if (obj->wake==NULL && *(taglist+1))
				catch( PWCreate((PWObject)obj,&obj->wake,PW_TYPE_LOOSE_ITEM,
								PW_POSITION_LEFT_OF, obj->doit?obj->doit:(obj->help?obj->help:(obj->info?obj->info:obj->title)),
								PW_LOOSE_TEXT, "wake",
/*								  PW_LOOSE_KEYPRESS, KEYSTROKE_WAKE,*/
								PW_LOOSE_KEYPRESS, 237,
								PW_LOOSE_ACTION_HIT, *(taglist+1),
								PW_LOOSE_CHANGE_STATUS, FALSE,
								PW_SCALE_FACTOR, 0,
								NULL) );
			break;

		case PW_OUTLINE_HELP:
			if (!obj->help)
				catch( PWCreate((PWObject)obj,&obj->wake,PW_TYPE_LOOSE_ITEM,
								PW_POSITION_LEFT_OF, obj->info?obj->info:obj->title,
								PW_LOOSE_TEXT, "help",
								PW_LOOSE_KEYPRESS, 0xe8,	/* F1 */
								PW_LOOSE_ACTION_HIT, help,
								PW_LOOSE_CHANGE_STATUS, FALSE,
								PW_SCALE_FACTOR, 0,
								PW_OBJECT_AUXILIARY, obj,
								NULL) );
			break;
		case PW_OUTLINE_ACTION_HELP:
			if (!obj->help)
			{
				static const int list[2]={PW_OUTLINE_HELP,NUL};
				catch( Change(object,list) );
			}
			catch( PWChange(obj->help,PW_LOOSE_ACTION_HIT,(*(taglist+1)),NULL) );
			break;
		case PW_OUTLINE_HELP_FILE:
			STRCopySafe((char *)(*(taglist+1)),obj->helpfile,IO_MAXFULLNAME);
			obj->helppos[0]='\0';
			break;
		case PW_OUTLINE_HELP_DIRECTORY:
			STRCopySafe((char *)(*(taglist+1)),obj->helpdir,IO_MAXDIRECTORY);
			break;
		case PW_OUTLINE_HELP_POSITION:
			STRCopySafe((char *)(*(taglist+1)),obj->helppos,IO_MAXFULLNAME);
			break;

		case PW_OUTLINE_TITLE_TEXT:
		case PW_TITLE_TEXT:
			catch( PWChange(obj->title,PW_TITLE_TEXT,(*(taglist+1)),NULL) );
			break;

		default :
			if (*taglist&(~PW_TAGMASK_SKIP))
				unrecognised_tag=TRUE;
			break;
		}
		taglist=PWSkipTag(taglist);
	}
	if (unrecognised_tag) err=ERR_IPAR;
	return err;
}

PWTypeDef init={
	PW_TYPEDEF_FLAG, PW_TYPE_OUTLINE, PW_TYPE_EVENT_REGION,
	&Create, &Change, &Query, &Remove, NULL, NULL, NULL, NULL,
	NULL, NULL, NULL
};
